@@ -5,29 +5,32 @@ The tool can be used in two ways:
 2. To launch a daemon which will log usage over time.  This can then later be queried
    to provide simple usage statistics.
 """
-import argparse
-import ast
-import atexit
 import os
-import random
-import signal
-import subprocess
+import re
+import ast
 import sys
 import time
-from collections import defaultdict
-from datetime import datetime
-from pathlib import Path
-from typing import Dict, List
+import atexit
+import signal
+import argparse
 import functools
+import subprocess
+import humanize
+import humanfriendly as hf
+from pathlib import Path
+from datetime import datetime
+from collections import defaultdict
 
-import colored
 import numpy as np
+import colored
 import seaborn as sns
-
 from beartype import beartype
+from beartype.cave import NoneType
+
 
 # SLURM states which indicate that the node is not available for submitting jobs
 INACCESSIBLE = {"drain*", "down*", "drng", "drain", "down"}
+INTERACTIVE_CMDS = {"bash", "zsh", "sh"}
 
 
 class Daemon:
@@ -205,7 +208,7 @@ class GPUStatDaemon(Daemon):
                 themselves.
         """
         if not Path(log_path).exists():
-            raise ValueError(f"No historical log found.  Did you start the daemon?")
+            raise ValueError("No historical log found.  Did you start the daemon?")
         with open(log_path, "r") as f:
             rows = f.read().splitlines()
         data = []
@@ -330,27 +333,36 @@ def parse_node_names(node_str):
     return names
 
 
-def parse_cmd(cmd):
-    """Parse the output of a shell command into a list of strings, one per line of output.
+def parse_cmd(cmd, split=True):
+    """Parse the output of a shell command...
+     and if split set to true: split into a list of strings, one per line of output.
 
     Args:
         cmd (str): the shell command to be executed.
-
+        split (bool): whether to split the output per line
     Returns:
         (list[str]): the strings from each output line.
     """
     output = subprocess.check_output(cmd, shell=True).decode("utf-8")
-    return [x for x in output.split("\n") if x]
+    if split:
+        output = [x for x in output.split("\n") if x]
+    return output
 
 
 @beartype
-def node_states() -> Dict[str, str]:
+def node_states(partition: (str, NoneType) = None) -> dict:
     """Query SLURM for the state of each managed node.
 
+    Args:
+        partition: the partition/queue (or multiple, comma separated) of interest.
+            By default None, which queries all available partitions.
+
     Returns:
         a mapping between node names and SLURM states.
     """
     cmd = "sinfo --noheader"
+    if partition:
+        cmd += f" --partition={partition}"
     rows = parse_cmd(cmd)
     states = {}
     for row in rows:
@@ -363,7 +375,7 @@ def node_states() -> Dict[str, str]:
 
 @beartype
 @functools.lru_cache(maxsize=64, typed=True)
-def occupancy_stats_for_node(node: str) -> Dict[str, int]:
+def occupancy_stats_for_node(node: str) -> dict:
     """Query SLURM for the occupancy of a given node.
 
     Args:
@@ -391,35 +403,53 @@ def occupancy_stats_for_node(node: str) -> Dict[str, int]:
     occupancy = {}
     for metric, alloc_val in metrics["AllocTRES"].items():
         cfg_val = metrics["CfgTRES"][metric]
+        if metric == "mem":
+            # SLURM appears to sometimes misformat large numbers, producing summary strings
+            # like 68G/257669M, rather than 68G/258G. The humanfriendly library provides
+            # a more reliable number parser, and the humanize library provides a nice
+            # formatter.
+            alloc_val = humanize.naturalsize(hf.parse_size(alloc_val), format="%d")
+            cfg_val = humanize.naturalsize(hf.parse_size(cfg_val), format="%d")
         occupancy[metric] = f"{alloc_val}/{cfg_val}"
     return occupancy
 
 
 @beartype
-def parse_all_gpus(default_gpus: int = 4) -> Dict[str, List]:
+def parse_all_gpus(partition: (str, NoneType) = None,
+                   default_gpus: int = 4,
+                   default_gpu_name: str = "NONAME_GPU") -> dict:
     """Query SLURM for the number and types of GPUs under management.
 
     Args:
+        partition: the partition/queue (or multiple, comma separated) of interest.
+            By default None, which queries all available partitions.
         default_gpus: The number of GPUs estimated for nodes that have incomplete SLURM
             meta data.
+        default_gpu_name: The name of the GPU for nodes that have incomplete SLURM meta
+        data.
 
     Returns:
         a mapping between node names and a list of the GPUs that they have available.
     """
-    cmd = "sinfo -o '%50N|%30G' --noheader"
+    cmd = "sinfo -o '%1000N|%1000G' --noheader"
+    if partition:
+        cmd += f" --partition={partition}"
     rows = parse_cmd(cmd)
     resources = defaultdict(list)
+
+    # Debug the regular expression below at
+    # https://regex101.com/r/RHYM8Z/3
+    p = re.compile(r'gpu:(?:(\w*):)?(\d*)(?:\(\S*\))?\s*')
+
     for row in rows:
         node_str, resource_strs = row.split("|")
         for resource_str in resource_strs.split(","):
             if not resource_str.startswith("gpu"):
                 continue
-            tokens = resource_str.strip().split(":")
+            match = p.search(resource_str)
+            gpu_type = match.group(1) if match.group(1) is not None else default_gpu_name
             # if the number of GPUs is not specified, we assume it is `default_gpus`
-            import ipdb; ipdb.set_trace()
-            if tokens[2] == "":
-                tokens[2] = default_gpus
-            gpu_type, gpu_count = tokens[1], int(tokens[2])
+            gpu_count = int(match.group(2)) if match.group(2) != "" else default_gpus
             node_names = parse_node_names(node_str)
             for name in node_names:
                 resources[name].append({"type": gpu_type, "count": gpu_count})
@@ -427,7 +457,7 @@ def parse_all_gpus(default_gpus: int = 4) -> Dict[str, List]:
 
 
 @beartype
-def resource_by_type(resources: Dict[str, Dict]) -> Dict[str, Dict]:
+def resource_by_type(resources: dict) -> dict:
     """Determine the cluster capacity by gpu type
 
     Args:
@@ -444,7 +474,7 @@ def resource_by_type(resources: Dict[str, Dict]) -> Dict[str, Dict]:
 
 
 @beartype
-def summary_by_type(resources: Dict[str, Dict], tag: str):
+def summary_by_type(resources: dict, tag: str):
     """Print out out a summary of cluster resources, organised by gpu type.
 
     Args:
@@ -462,7 +492,7 @@ def summary_by_type(resources: Dict[str, Dict], tag: str):
 
 
 @beartype
-def summary(mode: str, resources: Dict[str, Dict] = None, states: Dict[str, List] = None):
+def summary(mode: str, resources: dict = None, states: dict = None):
     """Generate a printed summary of the cluster resources.
 
     Args:
@@ -485,7 +515,7 @@ def summary(mode: str, resources: Dict[str, Dict] = None, states: Dict[str, List
 
 
 @beartype
-def gpu_usage(resources: Dict[str, Dict]) -> Dict[str, Dict]:
+def gpu_usage(resources: dict, partition: (str, NoneType) = None) -> dict:
     """Build a data structure of the cluster resource usage, organised by user.
 
     Args:
@@ -494,21 +524,28 @@ def gpu_usage(resources: Dict[str, Dict]) -> Dict[str, Dict]:
     Returns:
         (dict): a summary of resources organised by user (and also by node name).
     """
-    cmd = "squeue -O tres-per-node,nodelist,username --noheader"
+    cmd = "squeue -O tres-per-node:100,nodelist:100,username:100,jobid:100 --noheader"
+    if partition:
+        cmd += f" --partition={partition}"
+    detailed_job_cmd = "scontrol show jobid -dd %s"
     rows = parse_cmd(cmd)
     usage = defaultdict(dict)
     for row in rows:
         tokens = row.split()
         # ignore pending jobs
-        if len(tokens) < 3 or not tokens[0].startswith("gpu"):
+        if len(tokens) < 4 or not tokens[0].startswith("gpu"):
             continue
-        gpu_count_str, node_str, user = tokens
+        gpu_count_str, node_str, user, jobid = tokens
         gpu_count_tokens = gpu_count_str.split(":")
         num_gpus = int(gpu_count_tokens[-1])
         if len(gpu_count_tokens) == 2:
             gpu_type = None
         elif len(gpu_count_tokens) == 3:
             gpu_type = gpu_count_tokens[1]
+        # get detailed job information, to check if using bash
+        detailed_output = parse_cmd(detailed_job_cmd % jobid, split=False)
+        is_bash = any([f'Command={x}\n' in detailed_output for x in INTERACTIVE_CMDS])
+        num_bash_gpus = num_gpus * is_bash
         node_names = parse_node_names(node_str)
         for node_name in node_names:
             # If a node still has jobs running but is draining, it will not be present
@@ -518,22 +555,30 @@ def gpu_usage(resources: Dict[str, Dict]) -> Dict[str, Dict]:
             node_gpu_types = [x["type"] for x in resources[node_name]]
             if gpu_type is None:
                 if len(node_gpu_types) != 1:
-                    gpu_type = random.choice(node_gpu_types)
+                    gpu_type = sorted(
+                        resources[node_name],
+                        key=lambda k: k['count'],
+                        reverse=True
+                    )[0]['type']
                     msg = (f"cannot determine node gpu type for {user} on {node_name}"
-                           f" (guesssing {gpu_type})")
+                           f" (guessing {gpu_type})")
                     print(f"WARNING >>> {msg}")
                 else:
                     gpu_type = node_gpu_types[0]
             if gpu_type in usage[user]:
-                usage[user][gpu_type][node_name] += num_gpus
+                usage[user][gpu_type][node_name]['n_gpu'] += num_gpus
+                usage[user][gpu_type][node_name]['bash_gpu'] += num_bash_gpus
+
             else:
-                usage[user][gpu_type] = defaultdict(lambda: 0)
-                usage[user][gpu_type][node_name] += num_gpus
+                usage[user][gpu_type] = defaultdict(lambda: {'n_gpu': 0, 'bash_gpu': 0})
+                usage[user][gpu_type][node_name]['n_gpu'] += num_gpus
+                usage[user][gpu_type][node_name]['bash_gpu'] += num_bash_gpus
+
     return usage
 
 
 @beartype
-def in_use(resources: Dict[str, Dict] = None):
+def in_use(resources: dict = None, partition: (str, NoneType) = None):
     """Print a short summary of the resources that are currently used by each user.
 
     Args:
@@ -541,21 +586,27 @@ def in_use(resources: Dict[str, Dict] = None):
     """
     if not resources:
         resources = parse_all_gpus()
-    usage = gpu_usage(resources)
+    usage = gpu_usage(resources, partition=partition)
     aggregates = {}
     for user, subdict in usage.items():
-        aggregates[user] = {key: sum(val.values()) for key, val in subdict.items()}
+        aggregates[user] = {}
+        aggregates[user]['n_gpu'] = {key: sum([x['n_gpu'] for x in val.values()])
+                                     for key, val in subdict.items()}
+        aggregates[user]['bash_gpu'] = {key: sum([x['bash_gpu'] for x in val.values()])
+                                        for key, val in subdict.items()}
     print("Usage by user:")
-    for user, subdict in sorted(aggregates.items(), key=lambda x: sum(x[1].values())):
-        total = f"total: {str(sum(subdict.values())):2s}"
-        summary_str = ", ".join([f"{key}: {val}" for key, val in subdict.items()])
+    for user, subdict in sorted(aggregates.items(),
+                                key=lambda x: sum(x[1]['n_gpu'].values())):
+        total = (f"total: {str(sum(subdict['n_gpu'].values())):2s} "
+                 f"(interactive: {str(sum(subdict['bash_gpu'].values())):2s})")
+        summary_str = ", ".join([f"{key}: {val}" for key, val in subdict['n_gpu'].items()])
         print(f"{user:10s} [{total}] {summary_str}")
 
 
 @beartype
 def available(
-        resources: Dict[str, Dict] = None,
-        states: Dict[str, str] = None,
+        resources: dict = None,
+        states: dict = None,
         verbose: bool = False,
 ):
     """Print a short summary of resources available on the cluster.
@@ -563,6 +614,7 @@ def available(
     Args:
         resources: a summary of cluster resources, organised by node name.
         states: a mapping between node names and SLURM states.
+        verbose: whether to output a more verbose summary of the cluster state.
 
     NOTES: Some systems allow users to share GPUs.  The logic below amounts to a
     conservative estimate of how many GPUs are available.  The algorithm is:
@@ -583,7 +635,7 @@ def available(
             for node_name, user_gpu_count in node_dicts.items():
                 resource_idx = [x["type"] for x in res[node_name]].index(gpu_type)
                 count = res[node_name][resource_idx]["count"]
-                count = max(count - user_gpu_count, 0)
+                count = max(count - user_gpu_count['n_gpu'], 0)
                 res[node_name][resource_idx]["count"] = count
     by_type = resource_by_type(res)
     total = sum(x["count"] for sublist in by_type.values() for x in sublist)
@@ -592,7 +644,7 @@ def available(
         gpu_count = sum(x["count"] for x in counts_for_gpu_type)
         tail = ""
         if verbose:
-            summary = []
+            summary_strs = []
             for x in counts_for_gpu_type:
                 node, count = x["node"], x["count"]
                 if count:
@@ -600,16 +652,20 @@ def available(
                     users = [user for user in usage if node in usage[user].get(key, [])]
                     details = [f"{key}: {val}" for key, val in sorted(occupancy.items())]
                     details = f"[{', '.join(details)}] [{','.join(users)}]"
-                    summary.append(f"\n -> {node}: {count} {key} {details}")
-            tail = " ".join(summary)
+                    summary_strs.append(f"\n -> {node}: {count} {key} {details}")
+            tail = " ".join(summary_strs)
         print(f"{key}: {gpu_count} available {tail}")
 
 
 @beartype
-def all_info(color: int, verbose: bool):
+def all_info(color: int, verbose: bool, partition: (str, NoneType) = None):
     """Print a collection of summaries about SLURM gpu usage, including: all nodes
     managed by the cluster, nodes that are currently accesible and gpu usage for each
     active user.
+
+    Args:
+        partition: the partition/queue (or multiple, comma separated) of interest.
+            By default None, which queries all available partitions.
     """
     divider, slurm_str = "---------------------------------", "SLURM"
     if color:
@@ -619,12 +675,12 @@ def all_info(color: int, verbose: bool):
     print(divider)
     print(f"Under {slurm_str} management")
     print(divider)
-    resources = parse_all_gpus()
-    states = node_states()
+    resources = parse_all_gpus(partition=partition)
+    states = node_states(partition=partition)
     for mode in ("up", "accessible"):
         summary(mode=mode, resources=resources, states=states)
         print(divider)
-    in_use(resources)
+    in_use(resources, partition=partition)
     print(divider)
     available(resources=resources, states=states, verbose=verbose)
     print(divider)
@@ -639,6 +695,9 @@ def main():
                               "provide statistics from historical data (provided that the"
                               "logging daemon has been running). 'daemon-start' and"
                               "'daemon-stop' will start and stop the daemon, resp."))
+    parser.add_argument("-p", "--partition", default=None,
+                        help=("the partition/queue (or multiple, comma separated) of"
+                              " interest. By default set to all available partitions."))
     parser.add_argument("--log_path",
                         default=Path.home() / "data/daemons/logs/slurm_gpustat.log",
                         help="the location where daemon log files will be stored")
@@ -653,7 +712,7 @@ def main():
     args = parser.parse_args()
 
     if args.action == "current":
-        all_info(color=args.color, verbose=args.verbose)
+        all_info(color=args.color, verbose=args.verbose, partition=args.partition)
     elif args.action == "history":
         data = GPUStatDaemon.deserialize_usage(args.log_path)
         historical_summary(data)
